"""Auto-assessment library for Python"""

from . import utils
from .annotations import *
from .complexity import TimeComplexityChecker
from .debug import *
from .execution import *
from .reference import *
from .student import *
from .version import __version__
